/**
 *  @file iccvcgtchg.c
 *  @author tenk*
 *  @date   2014-10-10
 *  @note
 *      Boost Software License, Version 1.0
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define PEEK_U8(b)      (*(unsigned char*)(b))
#define PEEK_U16(b)     ((PEEK_U8(b) << 8) | PEEK_U8((char*)(b)+1))
#define PEEK_U32(b)     ((PEEK_U16(b) << 16) | PEEK_U16((char*)(b)+2))
#define POKE_U16(b,d)   ((*(unsigned char*)(b) = ((d) >> 8)), (*((unsigned char*)(b)+1) = (d)))

struct Options {
    unsigned int    rgb[3];
    int             print_flag;
    char const*     src_str;
    char const*     dst_str;
};
typedef struct Options Options;
void options_init(Options* options);
int  options_scan(Options* options, char const* src);

struct Vcgt {
    char*               data;
    size_t              data_size;
    char*               vcgt;
    unsigned short      plane;
    unsigned short      count;
    unsigned short      bytes;
    unsigned short*     rg;
    unsigned short*     gg;
    unsigned short*     bg;
};
typedef struct Vcgt Vcgt;
int  vcgt_find(Vcgt* vcgt, char* data, size_t size);
void vcgt_disp(Vcgt* vcgt, char const* fname);
void vcgt_setGammaTable(Vcgt* vcgt, int r, int g, int b);
int  icc_replaceStr(char* data, size_t data_size, char const *src, char const *dst);

char* file_load(char const* fname, size_t *sizePtr);
int   file_save(char const* fname, void* data, size_t size);
void  file_free(void* data);
void* MemMem(const void *src, size_t len, const void *ptn, size_t ptnlen);


//----------------------------------------------------------------------------

int  conv(char const* srcname, char const* dstname, Options* options);

int usage()
{
    printf("usage> iccvcgtchg.exe [-opts] src-fname [dst-fname]\n"
           " opts:\n"
           "  -r<NNNN>  red   value (0000-FFFF)\n"
           "  -g<NNNN>  green value (0000-FFFF)\n"
           "  -b<NNNN>  blue  value (0000-FFFF)\n"
           "  -p        print gamma table\n"
           "  -s<STR>   replace string source.\n"
           "  -d<STR>   replace string destnation.\n"
    );
    return 1;
}


int main(int argc, char* argv[])
{
    Options options;
    int     i  = 0;
    int     rc = 0;
    char*   srcname = NULL;
    char*   dstname = NULL;

    if (argc < 2)
        return usage();

    options_init(&options);

    for (i = 1; i < argc; ++i) {
        char* p = argv[i];
        if (*p == '-') {
            rc = options_scan(&options, p);
            if (!rc)
                return 1;
        } else if (!srcname) {
            srcname = p;
        } else if (!dstname) {
            dstname = p;
        } else {
            printf("too many filenames.\n");
            return 1;
        }
    }
    if (!srcname) {
        printf("no filename\n");
        return 1;
    }

    rc = conv(srcname, dstname, &options);
    return !rc;
}


int conv(char const* srcname, char const* dstname, Options* options)
{
    Vcgt   vcgt;
    size_t size = 0;
    char*  data = file_load(srcname, &size);
    if (!data)
        return 0;
    if (vcgt_find(&vcgt, data, size) == 0) {
        printf("%s : unkown format\n", srcname);
        return 0;
    }
    if (options->src_str && options->dst_str) {
        if (icc_replaceStr(data, size, options->src_str, options->dst_str) == 0)
            return 0;
    }
    if (dstname) {
        vcgt_setGammaTable(&vcgt, options->rgb[0], options->rgb[1], options->rgb[2]);
        file_save(dstname, data, size);
    }
    if (options->print_flag)
        vcgt_disp(&vcgt, srcname);
    file_free(data);
    return 1;
}


//----------------------------------------------------------------------------

void options_init(Options* options)
{
    options->rgb[0]     = 0xfeff;
    options->rgb[1]     = 0xfeff;
    options->rgb[2]     = 0xfeff;
    options->print_flag = 0;
    options->src_str    = NULL;
    options->dst_str    = NULL;
}

static int options_strToRgb(Options* options, int idx, char const* s, char const* src)
{
    int c = strtoul(s, NULL, 16);
    if (c < 0 || c > 0xffff) {
        printf("%s : out of range\n", src);
        return 0;
    }
    options->rgb[idx] = c;
    return 1;
}

int options_scan(Options* options, char const* src)
{
    char const* s = src+1;
    int c = *s++;

    switch (c) {
    case '\0':
    case '?':
        usage();
        break;
    case 'p':
        options->print_flag = 1;
        break;
    case 'r':
        if (options_strToRgb(options, 0, s, src) == 0)
            return 0;
        break;
    case 'g':
        if (options_strToRgb(options, 1, s, src) == 0)
            return 0;
        break;
    case 'b':
        if (options_strToRgb(options, 2, s, src) == 0)
            return 0;
        break;
    case 's':
        options->src_str = s;
        break;
    case 'd':
        options->dst_str = s;
        break;
    default:
        printf("unkown option %s\n", src);
        return 0;
    }
    return 1;
}


//----------------------------------------------------------------------------

char* file_load(char const* fname, size_t *sizePtr)
{
    long  size = 0;
    char* data = NULL;
    FILE* fp   = fopen(fname, "rb");
    if (!fp) {
        printf("%s : file not open\n", fname);
        return NULL;
    }

    if (fseek(fp, 0, SEEK_END) != 0) {
        printf("%s : seek error\n", fname);
        return NULL;
    }
    size = ftell(fp);
    if ((unsigned long)size > 0x10000000L || size == 0) {
        fclose(fp);
        printf("%s : bad file size(%ld bytes)\n", fname, size);
        return NULL;
    }
    data = (char*)calloc(1, size);
    if (!data) {
        printf("%s : not enough memory(%ld bytes)\n", fname, size);
        fclose(fp);
        return NULL;
    }
    fseek(fp, 0, SEEK_SET);
    if (fread(data, 1, size, fp) != size) {
        printf("%s : read error\n", fname);
        fclose(fp);
        free(data);
        return NULL;
    }
    fclose(fp);
    *sizePtr = size;
    return data;
}

int file_save(char const* fname, void* data, size_t size)
{
    FILE* fp = fopen(fname, "wb");
    if (!fp) {
        printf("%s : file not open\n", fname);
        return 0;
    }
    if (fwrite(data, 1, size, fp) != size) {
        printf("%s : write error\n", fname);
        fclose(fp);
        return 0;
    }
    fclose(fp);
    return 1;
}

void file_free(void* data)
{
    free(data);
}


//----------------------------------------------------------------------------

void* MemMem(const void *src, size_t len, const void *ptn, size_t ptnlen)
{
    unsigned char const *s;
    unsigned char const *se;
    unsigned char const *p;
    unsigned char const *pe;
    unsigned             k;
    unsigned             c;

    if (ptnlen == 0 || ptn == 0)
        return (void*)src;
    if (len == 0 || src == 0)
        return NULL;

    s  = (unsigned char const*)src;
    se = (unsigned char const*)src + len;
    p  = (unsigned char const*)ptn;
    pe = p + ptnlen;
    k  = *p++;

    while (s < se) {
        c = *s++;
        if (k == c) {
            unsigned char const *l, *r;
            l = s;
            r = p;
            c = 0;
            while (r != pe) {
                c = *l++;
                c = c - *r++;
                if (c != 0)
                    break;
            }
            if (c == 0)
                return (void*)(s - 1);
        }
    }
    return NULL;
}

//----------------------------------------------------------------------------

int  vcgt_find(Vcgt* vcgt, char* data, size_t size)
{
    char*  fnd  = data;
    size_t rest = size;
    do {
        fnd = (char*)MemMem(fnd, rest, "vcgt", 4);
        if (!fnd) {
            printf("not found (or unsuported) vcgt\n");
            return 0;
        }
        fnd  += 4;
        rest  = data + size - fnd;
        if (rest < 14+3*2*256) {
            printf("unsuported vcgt\n");
            return 0;
        }
        if (PEEK_U32(fnd + 0) != 0 || PEEK_U32(fnd + 4) != 0) {
            continue;
        }
        vcgt->data  = data;
        vcgt->data_size = size;
        vcgt->vcgt  = fnd;
        vcgt->plane = PEEK_U16(fnd+ 8); // 3
        vcgt->count = PEEK_U16(fnd+10); // 256
        vcgt->bytes = PEEK_U16(fnd+12); // 2
        vcgt->rg    = (unsigned short*)(fnd + 14 + 0 * 2 * vcgt->count);
        vcgt->gg    = (unsigned short*)(fnd + 14 + 1 * 2 * vcgt->count);
        vcgt->bg    = (unsigned short*)(fnd + 14 + 2 * 2 * vcgt->count);
    } while (vcgt->plane != 3 || vcgt->count != 256 || vcgt->bytes != 2);
    return 1;
}

void vcgt_disp(Vcgt* vcgt, char const* fname)
{
    int i;
    printf("%s\n", fname);
    for (i = 0; i < vcgt->count; ++i) {
        printf("\t[%#04x] %#06x, %#06x, %#06x\n", i, PEEK_U16(vcgt->rg+i), PEEK_U16(vcgt->gg+i), PEEK_U16(vcgt->bg+i));
    }
}

void vcgt_setGammaTable(Vcgt* vcgt, int rg, int gg, int bg)
{
    int i;
    int count = vcgt->count;
    for (i = 0; i < count; ++i) {
        int r = i * rg / (count - 1);
        int g = i * gg / (count - 1);
        int b = i * bg / (count - 1);
        POKE_U16(vcgt->rg+i, r);
        POKE_U16(vcgt->gg+i, g);
        POKE_U16(vcgt->bg+i, b);
    }
}

int icc_replaceStr(char* data, size_t data_size, char const *src, char const *dst)
{
    enum { MAX_SIZE = 100 };
    unsigned short  src2[MAX_SIZE+1];
    unsigned short  dst2[MAX_SIZE+1];
    int             i = 0;
    char*           p = NULL;
    size_t          len = strlen(src);

    if (len >= MAX_SIZE) {
        printf("%s : too long replace-source-string.\n",src);
        return 0;
    }
    if (len != strlen(dst)) {
        printf("%s , %s : length is different.\n",src, dst);
        return 0;
    }
    memset(src2, 0, sizeof src2);
    memset(dst2, 0, sizeof dst2);
    for (i = 0; i < len; ++i) {
        POKE_U16(src2+i, src[i]);
        POKE_U16(dst2+i, dst[i]);
    }
    p = MemMem(data, data_size, src2, len*2);
    if (!p) {
        printf("%s : not found.\n", src);
        return 0;
    }
    memcpy(p, dst2, len*2);
    return 1;
}
